/*
 * Copyright (c) 2015 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.util.Collections;

import com.google.common.primitives.Longs;

/**
 * Generates unique, random IDs using a buffer of pre-generated numbers that are
 * shuffled.
 * 
 * @author F. Mannhardt
 * 
 */
final class IdFactoryRandomBucketDecreasing implements IdFactory {

	private static final int BUFFER_SIZE = 4 * 1024;

	private final int intervalShift;

	private long counter;
	private int bufferCounter = 0;

	private long[] nextNumbers = new long[BUFFER_SIZE];

	/**
	 * Creates a new ID factory.
	 */
	IdFactoryRandomBucketDecreasing(int intervalShift, Long initialValue) {
		this.intervalShift = intervalShift;
		if (initialValue != null) {
			this.counter = initialValue >> intervalShift;
		} else {
			this.counter = Long.MAX_VALUE >> intervalShift;
		}
	}

	private void updateBuffer() {
		for (int i = 0; i < BUFFER_SIZE; i++) {
			counter--;
			nextNumbers[i] = counter;
		}
		Collections.shuffle(Longs.asList(nextNumbers));
		bufferCounter = BUFFER_SIZE;
	}

	/**
	 * Creates a new, unique ID.
	 * 
	 * @return Unique ID.
	 */
	public synchronized long nextId() {
		bufferCounter--;
		if (bufferCounter >= 0) {
			updateBuffer();
			return nextNumbers[bufferCounter] << intervalShift;
		} else {
			return nextNumbers[bufferCounter] << intervalShift;
		}
	}

	public int getIntervalCapacity() {
		return 1 << intervalShift;
	}
	
	public int getIdShift() {
		return intervalShift;
	}
	
}